@@ -29,6 +29,9 @@ from twisted.python import components, filepath, log
 from twisted.internet import abstract, interfaces
 from twisted.python.util import InsensitiveDict
 from twisted.python.runtime import platformType
+from twisted.python.url import URL
+from twisted.python.versions import Version
+from twisted.python.deprecate import deprecated
 
 if _PY3:
     from urllib.parse import quote, unquote
@@ -61,22 +64,37 @@ class Data(resource.Resource):
     render_HEAD = render_GET
 
 
-def addSlash(request):
-    qs = ''
-    qindex = request.uri.find('?')
-    if qindex != -1:
-        qs = request.uri[qindex:]
 
-    return "http%s://%s%s/%s" % (
-        request.isSecure() and 's' or '',
-        request.getHeader("host"),
-        (request.uri.split('?')[0]),
-        qs)
+@deprecated(Version("Twisted", 16, 0, 0))
+def addSlash(request):
+    """
+    Add a trailing slash to C{request}'s URI. Deprecated, do not use.
+    """
+    return _addSlash(request)
+
+
+
+def _addSlash(request):
+    """
+    Add a trailing slash to C{request}'s URI.
+
+    @param request: The incoming request to add the ending slash to.
+    @type request: An object conforming to L{twisted.web.iweb.IRequest}
+
+    @return: A URI with a trailing slash, with query and fragment preserved.
+    @rtype: L{bytes}
+    """
+    url = URL.fromText(request.uri.decode('ascii'))
+    # Add an empty path segment at the end, so that it adds a trailing slash
+    url = url.replace(path=list(url.path) + [u""])
+    return url.asText().encode('ascii')
+
+
 
 class Redirect(resource.Resource):
     def __init__(self, request):
         resource.Resource.__init__(self)
-        self.url = addSlash(request)
+        self.url = _addSlash(request)
 
     def render(self, request):
         return redirectTo(self.url, request)
@@ -610,7 +628,7 @@ class File(resource.Resource, filepath.FilePath):
 
 
     def redirect(self, request):
-        return redirectTo(addSlash(request), request)
+        return redirectTo(_addSlash(request), request)
 
 
     def listNames(self):
